@@ -8,6 +8,7 @@ import sys
 from attrdict import AttrDict
 import attrdict
 from getpass import getpass
+from queue import Queue
 
 class Bot:
 
@@ -23,17 +24,6 @@ class Bot:
         response = random.choice(responses)
         return response
 
-    def add_command_line_arg(self, *args, **kwargs):
-        self.parser.add_argument(*args, **kwargs)
-        self.args = self.parser.parse_args()
-        if "name" in kwargs:
-            return self.args.kwargs["name"]
-        else:
-            return self.args
-
-    def send_command(self, type = "send", data = {}):
-        self.conn.send(json.dumps({'type' : type, 'data' : data}))
-
     def __init__(self, nick, **kwargs):
         vars = ["userEmail",
             "userPassword",
@@ -60,12 +50,14 @@ class Bot:
         self.normname = re.sub(r"\s+", "", self.nick)
         self.owner = "nobody"
         self.handlers = {}
-        self.copies = []
+        self.children = []
         self.important = False
         self.APITimeout = 10
         self.advanced = False
         self.sendable = False
-        self.copyThreads = []
+        self.childThreads = []
+        self.parents = []
+        self.parentThreads = []
         self.conn = None
         self.human = False
         self._regexes = {}
@@ -90,96 +82,46 @@ class Bot:
             if not self._copy:
                 log(f"Debug: {str(self.args.test)}", _copy = self._copy)
 
-    def connect(self):
-        self.conn = websocket.create_connection(f'wss://euphoria.io/room/{self.room}/ws{"?h=0" if not self.human else "?h=1"}', cookie = self.cookie, enable_multithread = True)
-        if self.cookie is None:
-            self.cookie = self.conn.headers['set-cookie']
-        hello = AttrDict(json.loads(self.conn.recv()))
-        reply = AttrDict(json.loads(self.conn.recv()))
-        self._handle_ping(reply)
-        reply = AttrDict(json.loads(self.conn.recv()))
-        if reply.type == "snapshot-event":
-            self.set_nick(self.nick)
-        elif reply.type == "bounce-event":
-            self._handle_auth(self.password)
-            self.set_nick(self.nick)
-        else:
-            log(reply, error = True, _copy = self._copy)
-        log(f'connected to &{self.room}.', _copy = self._copy)
-        return hello
-
-    def send_msg(self, msgString, parent = None):
-        try:
-            if isinstance(parent, AttrDict):
-                self.send_command(data = {'content': msgString, 'parent': parent.data.id})
-            elif isinstance(parent, str):
-                self.send_command(data = {'content': msgString, 'parent': parent})
-            elif parent is None:
-                self.send_command(data = {'content': msgString})
+    def _wait_for(self, response):
+        reply = None
+        i = 0
+        waitlist = Queue()
+        for msgJSON in self.conn:
+            i += 1
+            msg = AttrDict(json.loads(msgJSON))
+            if msg.type == response:
+                reply = msg
+                break
             else:
-                raise BadMessageError(f'message not sent, because type of parent was {type(parent)}. \nParent printed: \n{parent}')
-            reply = None
-            i = 0
-            for msgJSON in self.conn:
-                i += 1
-                msg = AttrDict(json.loads(msgJSON))
-                if msg.type == "send-reply":
-                    reply = msg
-                    if isinstance(parent, AttrDict):
-                        log(f'Message sent: {reply} replying to: {parent.data.id} by {parent.data.sender.name}', _copy = self._copy)
-                    elif isinstance(parent, str):
-                        log(f'Message sent: {reply} replying to: {parent}', _copy = self._copy)
-                    elif parent is None:
-                        log(f'Message sent: {reply}', _copy = self._copy)
-                    return reply
-                if i > self.APITimeout:
-                    log("send-reply API response not recorded. reason: too many events before send-reply.", True, _copy = self._copy)
-                    return None
-        except BadMessageError:
-            pass
+                waitlist.put(msg)
+            if i > self.APITimeout:
+                log(f"{response} API response not recorded. reason: too many events before {response}.", error = True, _copy = self._copy)
+                reply = None
+                break
+        while not waitlist.empty():
+            self._handle(waitlist.get())
+        return reply
 
-    def restart(self, msg = None, _copyStart = False):
-        if self.conn is not None:
-            self.conn.close()
-            del self.conn
-            log("restarting...", _copy = self._copy)
-        else:
-            log("starting...", _copy = self._copy)
-        self.connect()
-        self.start()
+    def _recv(self):
+        return AttrDict(json.loads(self.conn.recv()))
 
-    def start(self, function = None, advanced = False):
-        if not advanced and not self.advanced:
-            self.set_handler('ping-event', self._handle_ping)
-            self.set_handler('send-event', self._handle_message)
-            self.set_handler('bounce-event', self._handle_auth)
+    def _handle(self, msg):
+        if msg.type == 'send-event':
+            if re.search(f'^!kill @{re.escape(self.normname)}$', msg.data.content) is not None:
+                self._handle_kill(msg)
+            if re.search(f'^!forcekill @{re.escape(self.normname)}$', msg.data.content) is not None:
+                self.kill()
+            if re.search(f'^!restart @{re.escape(self.normname)}$', msg.data.content) is not None:
+                self.restart(msg = msg)
+        if msg.type == 'error':
+            log(msg, error = True, _copy = self._copy)
+        elif msg.type in self.handlers:
+            self.handlers[msg.type](msg)
         else:
-            if self.function is None and function is not None and callable(function):
-                self.function = function
+            if callable(self.function):
+                self.function(msg)
             else:
-                log("No valid function passed to advanced bot's start function. Using filler function", _copy = self._copy)
-                self.function = lambda msg: None
-        try:
-            for msgJSON in self.conn:
-                msg = AttrDict(json.loads(msgJSON))
-                if msg.type == 'send-event':
-                    if re.search(f'^!kill @{re.escape(self.normname)}$', msg.data.content) is not None:
-                        self._handle_kill(msg)
-                    if re.search(f'^!forcekill @{re.escape(self.normname)}$', msg.data.content) is not None:
-                        self.kill()
-                    if re.search(f'^!restart @{re.escape(self.normname)}$', msg.data.content) is not None:
-                        self.restart(msg = msg)
-                if msg.type == 'error':
-                    log(msg, error = True, _copy = self._copy)
-                elif msg.type in self.handlers:
-                    self.handlers[msg.type](msg)
-                else:
-                    if callable(self.function):
-                        self.function(msg)
-                    else:
-                        log("Advanced start must be given a callable message handler function that takes an AttrDict as its argument.", error = True, _copy = self._copy)
-        except KilledError:
-            pass
+                log("Advanced start must be given a callable message handler function that takes an AttrDict as its argument.", error = True, _copy = self._copy)
 
     def _is_privileged(self, user):
         if "is_manager" in user or user.name == self.owner:
@@ -187,26 +129,6 @@ class Bot:
         else:
             return False
 
-    def set_regexes(self, regexes):
-        self._regexes = regexes
-        if self.help is None:
-            if f"^!help @{re.escape(self.normname)}$" in regexes:
-                self.help = regexes[f"^!help @{re.escape(self.normname)}$"]
-            else:
-                self.help = f"@{self.normname} is a bot made with Doctor Number Four's Python 3 bot library, DocLib (link: https://github.com/milovincent/DocLib) by @{self.normowner}.\nIt follows botrulez and does not have a custom !help message yet."
-        if self.ping is None:
-            if '^!ping$' in regexes:
-                self.ping = regexes['^!ping$']
-            elif f'^!ping @{re.escape(self.normname)}$' in regexes:
-                self.ping = regexes[f'^!ping @{re.escape(self.normname)}$']
-            else:
-                self.ping = "Pong!"
-        if self.killed is None:
-            if f"^!kill @{re.escape(self.normname)}$" in regexes:
-                self.killed = regexes[f"^!kill @{re.escape(self.normname)}$"]
-            else:
-                self.killed = "/me has died."
-
     def _handle_ping(self, msg):
         self.send_command(type = 'ping-reply', data = {'time': msg.data.time})
 
@@ -255,9 +177,9 @@ class Bot:
     def _handle_kill(self, msg):
         if self._is_privileged(msg.data.sender) or not self.important:
             self.send_msg(self.killed, msg)
-            for process in self.copyThreads:
+            for process in self.childThreads:
                 process.terminate()
-            for copy in self.copies:
+            for copy in self.children:
                 copy.send_msg("/me was killed via parricide")
                 copy.kill()
             self.kill()
@@ -266,26 +188,110 @@ class Bot:
 
     def _handle_auth(self, pw):
         self.send_command(type = 'auth', data = {'type': 'passcode', 'passcode': pw})
-        reply = None
-        i = 0
-        for msgJSON in self.conn:
-            i += 1
-            msg = AttrDict(json.loads(msgJSON))
-            if msg.type == "auth-reply":
-                reply = msg
-                if reply.data.success == True:
-                    log(f'Successfully logged into {self.room}.', _copy = self._copy)
-                else:
-                    log(f'Login unsuccessful. Reason: {reply.data.reason}', error = True, _copy = self._copy)
-                    self._handle_auth(getpass("Enter the correct password: "))
-                break
-            if i > self.APITimeout:
-                log("Auth Error: auth-reply API response not recorded.", error = True, _copy = self._copy)
-                break
+        reply = self._wait_for("auth-reply")
+        if reply is not None:
+            if reply.data.success == True:
+                log(f'Successfully logged into {self.room}.', _copy = self._copy)
+            else:
+                log(f'Login unsuccessful. Reason: {reply.data.reason}', error = True, _copy = self._copy)
+                self._handle_auth(getpass("Enter the correct password: "))
 
-    def _handle_other(self, msg):
-        if msg.type in self.handlers:
-            self.handlers[msg.type](msg)
+    def connect(self):
+        self.conn = websocket.create_connection(f'wss://euphoria.io/room/{self.room}/ws{"?h=0" if not self.human else "?h=1"}', cookie = self.cookie, enable_multithread = True)
+        if self.cookie is None:
+            self.cookie = self.conn.headers['set-cookie']
+        hello = self.recv()
+        reply = self.recv()
+        self._handle_ping(reply)
+        reply = self.recv()
+        if reply.type == "snapshot-event":
+            self.set_nick(self.nick)
+        elif reply.type == "bounce-event":
+            self._handle_auth(self.password)
+            self.set_nick(self.nick)
+        else:
+            log(reply, error = True, _copy = self._copy)
+        log(f'connected to &{self.room}.', _copy = self._copy)
+        return hello
+
+    def send_msg(self, msgString, parent = None):
+        try:
+            if isinstance(parent, AttrDict):
+                self.send_command(data = {'content': msgString, 'parent': parent.data.id})
+            elif isinstance(parent, str):
+                self.send_command(data = {'content': msgString, 'parent': parent})
+            elif parent is None:
+                self.send_command(data = {'content': msgString})
+            else:
+                raise BadMessageError(f'message not sent, because type of parent was {type(parent)}. \nParent printed: \n{parent}')
+            reply = self._wait_for("send-reply")
+            if isinstance(parent, AttrDict):
+                log(f'Message sent: {reply} replying to: {parent.data.id} by {parent.data.sender.name}', _copy = self._copy)
+            elif isinstance(parent, str):
+                log(f'Message sent: {reply} replying to: {parent}', _copy = self._copy)
+            elif parent is None:
+                log(f'Message sent: {reply}', _copy = self._copy)
+        except BadMessageError:
+            pass
+
+    def restart(self, msg = None, _copyStart = False):
+        if self.conn is not None:
+            self.conn.close()
+            del self.conn
+            log("restarting...", _copy = self._copy)
+        else:
+            log("starting...", _copy = self._copy)
+        self.connect()
+        self.start()
+
+    def start(self, function = None, advanced = False):
+        if not advanced and not self.advanced:
+            self.set_handler('ping-event', self._handle_ping)
+            self.set_handler('send-event', self._handle_message)
+            self.set_handler('bounce-event', self._handle_auth)
+        else:
+            if self.function is None and function is not None and callable(function):
+                self.function = function
+            else:
+                log("No valid function passed to advanced bot's start function. Using filler function", _copy = self._copy)
+                self.function = lambda msg: None
+        try:
+            for msgJSON in self.conn:
+                msg = AttrDict(json.loads(msgJSON))
+                self._handle(msg)
+        except KilledError:
+            pass
+
+    def add_command_line_arg(self, *args, **kwargs):
+        self.parser.add_argument(*args, **kwargs)
+        self.args = self.parser.parse_args()
+        if "name" in kwargs:
+            return self.args.kwargs["name"]
+        else:
+            return self.args
+
+    def send_command(self, type = "send", data = {}):
+        self.conn.send(json.dumps({'type' : type, 'data' : data}))
+
+    def set_regexes(self, regexes):
+        self._regexes = regexes
+        if self.help is None:
+            if f"^!help @{re.escape(self.normname)}$" in regexes:
+                self.help = regexes[f"^!help @{re.escape(self.normname)}$"]
+            else:
+                self.help = f"@{self.normname} is a bot made with Doctor Number Four's Python 3 bot library, DocLib (link: https://github.com/milovincent/DocLib) by @{self.normowner}.\nIt follows botrulez and does not have a custom !help message yet."
+        if self.ping is None:
+            if '^!ping$' in regexes:
+                self.ping = regexes['^!ping$']
+            elif f'^!ping @{re.escape(self.normname)}$' in regexes:
+                self.ping = regexes[f'^!ping @{re.escape(self.normname)}$']
+            else:
+                self.ping = "Pong!"
+        if self.killed is None:
+            if f"^!kill @{re.escape(self.normname)}$" in regexes:
+                self.killed = regexes[f"^!kill @{re.escape(self.normname)}$"]
+            else:
+                self.killed = "/me has died."
 
     def set_nick(self, nick):
         self.send_command(type = 'nick', data = {'name': nick})
@@ -296,17 +302,9 @@ class Bot:
 
     def get_userlist(self):
         self.send_command(type = 'who')
-        reply = None
-        i = 0
-        for msgJSON in self.conn:
-            i += 1
-            msg = AttrDict(json.loads(msgJSON))
-            if msg.type == "who-reply":
-                reply = msg
-                return reply.data.listing
-            if i > self.APITimeout:
-                log("who-reply API response not recorded. reason: too many events before who-reply.", error = True, _copy = self._copy)
-                return None
+        reply = self._wait_for("who-reply")
+        if reply is not None:
+            return reply.data.listing
 
     def move_to(self, roomName, password = "", _copyStart = False):
         self.room = roomName
@@ -346,15 +344,17 @@ class Bot:
         copy.set_handlers(self.handlers)
         copy.set_regexes(self._regexes)
         t = threading.Thread(target = copy.move_to)
-        self.copies.append(copy)
-        self.copyThreads.append(t)
+        self.children.append(copy)
+        copy.parents.append(self)
+        self.childThreads.append(t)
+        copy.parentThreads.append(threading.current_thread())
         t.run(roomName = roomName, password = password, _copyStart = True)
         t.join(0)
         return copy
 
     def login(self, email = None, password = None, setDefaults = False):
         if email is not None and password is not None:
-            self.send_command(type = 'login', data = {'namespace' : "email", 'id' : email, 'password' : password}})
+            self.send_command(type = 'login', data = {'namespace' : "email", 'id' : email, 'password' : password})
             if setDefaults:
                 self.userEmail = email
                 self.userPassword = password
@@ -388,26 +388,16 @@ class Bot:
         del copy.conn
         copy.connect()
         self.send_command(type = 'pm-initiate', data = {'user_id' : id})
-        to = None
-        i = 0
-        for msgJSON in copy.conn:
-            i += 1
-            msg = AttrDict(json.loads(msgJSON))
-            if msg.type == "pm-initiate-reply":
-                copy.conn.close()
-                to = msg
-                if to.data is not None:
-                    if bot is not None:
-                        bot.copy_to(f"pm:{to.data.pm_id}")
-                    else:
-                        self.copy_to("pm:{to.data.pm_id}")
+        to = self._wait_for("pm-initiate-reply")
+        if to is not None:
+            if to.data is not None:
+                if bot is not None:
+                    bot.copy_to(f"pm:{to.data.pm_id}")
                 else:
-                    log("Access Denied: This bot does not have permission to initiate PMs.", error = True, _copy = self._copy)
-                    log(f"reply: {msg}", error = True, _copy = self._copy)
-                return to
-            if i > self.APITimeout:
-                log("pm-initiate-reply API response not recorded. reason: too many events before pm-initiate-reply.", error = True, _copy = self._copy)
-                return None
+                    self.copy_to("pm:{to.data.pm_id}")
+            else:
+                log("Access Denied: This bot does not have permission to initiate PMs.", error = True, _copy = self._copy)
+        return to
 
 class BotError(Exception):
     pass
